@@ -109,12 +109,13 @@ class Workflow:
         if self.config.debug.save:
             assert self.output, "Saving debug output requires specifying an output path!"
 
-        # Setup initial stuff
-        self._starting_url = url
-        if isinstance(self._starting_url, str):
-            self._starting_url = {Workflow.SINGLE_TAB: self._starting_url}
-        if isinstance(list(self._starting_url.values())[0], str):
-            self._starting_url = {Workflow.SINGLE_TAB: self._starting_url}  # type: ignore
+        # Setup starting points
+        if isinstance(url, str):
+            url = {Workflow.SINGLE_TAB: url}
+        if isinstance(list(url.values())[0], str):
+            url = {Workflow.SINGLE_TAB: url}  # type: ignore
+        self._starting_url: Dict[str, Dict[str, str]] = url  # type: ignore
+
         self.reset()
 
     def __enter__(self):
@@ -202,6 +203,12 @@ class Workflow:
         if policy_stopped:
             self._execute_policy_result({tab: Abort() for tab in self.open_tabs})
 
+        # Reduce memory usage if keeping a lot of views
+        if self.config.scraping.history and not self.config.scraping.full_history:
+            for tab in self.open_tabs:
+                self.tab(tab)
+                self.history[-1] = self.history[-1].copy(no_snapshot=True)
+
         # Continue if there is still something in the policy
         return policy_stopped
 
@@ -223,59 +230,68 @@ class Workflow:
                 not self.config.scraping.all
                 and self.previous_policy_result
                 and tab not in self.previous_policy_result
-                and self.history[-1]
-                and self.history[-1].snapshot
+                and self.latest_view
+                and self.latest_view.snapshot
             ):
-                self.history.append(self.history[-1])
+                self.history.append(self.latest_view)
                 all_views[tab] = self.history[-1]
                 continue
 
-            # Ensure page is fully loaded
-            self.current_window.scraper.wait_until_loaded()
+            # Store View
+            all_views[tab] = self._get_new_view(tab, initial_action)
 
-            # Run postload callbacks
-            for cb in self.current_window.scraper.postload_callbacks:
-                cb()
+        return all_views
 
-            # Scrape the page
-            snapshot = self.current_window.scraper.scrape_current_page()
+    def _get_new_view(self, name: str, initial_action: Action) -> View:
+        # Ensure page is fully loaded
+        self.current_window.scraper.wait_until_loaded()
 
-            # Assemble basic list of actions
-            action_list: List[Action] = [Abort(), Refresh(), Navigate(), Wait()]
-            action_list += [Revert(step) for step in range(len(self.history))]
-            actions = Actions(action_list)
+        # Run postload callbacks
+        for cb in self.current_window.scraper.postload_callbacks:
+            cb()
 
-            # Setup metadata
+        # Scrape the page
+        snapshot = self.current_window.scraper.scrape_current_page()
+
+        # Assemble basic list of actions
+        action_list: List[Action] = [Abort(), Refresh(), Navigate(), Wait()]
+        action_list += [Revert(step) for step in range(len(self.history))]
+        actions = Actions(action_list)
+
+        # Setup metadata
+        metadata: Dict[str, Any] = {}
+        if self.config.scraping.history:
             if self.history and "next_action" in self.history[self.loop_idx - 1].metadata:
                 metadata = self.history[self.loop_idx - 1].metadata.copy()
                 metadata["previous_action"] = metadata["next_action"]
             else:
-                metadata = {}
                 metadata["previous_action"] = [initial_action]
-            metadata["next_action"] = None
+        metadata["next_action"] = None
 
-            # Create view
-            view = View(name=tab, snapshot=snapshot, actions=actions, metadata=metadata)
+        # Create view
+        view = View(name=name, snapshot=snapshot, actions=actions, metadata=metadata)
 
-            # Add state to history
-            if self.loop_idx < len(self.history):
-                self.history[self.loop_idx] = view
-            else:
-                self.history.append(view)
+        # Add state to history
+        if self.loop_idx < len(self.history):
+            self.history[self.loop_idx] = view
+        else:
+            self.history.append(view)
 
-            # Run element classifiers
-            view.actions.extend(self._run_element_classifiers(snapshot))
+        # Maintain only one level of history if required
+        if not self.config.scraping.history:
+            for i in range(len(self._history[self.current_tab]) - 1):
+                self._history[self.current_tab][i] = None
 
-            # Run page classifiers
-            for classifier in self.classifiers:
-                if classifier.enabled and classifier.callback and isinstance(classifier, ViewClassifier):
-                    result = classifier.callback(view)
-                    view.tags.update(result)
+        # Run element classifiers
+        view.actions.extend(self._run_element_classifiers(snapshot))
 
-            # Store views
-            all_views[tab] = view
+        # Run page classifiers
+        for classifier in self.classifiers:
+            if classifier.enabled and classifier.callback and isinstance(classifier, ViewClassifier):
+                result = classifier.callback(view)
+                view.tags.update(result)
 
-        return all_views
+        return view
 
     def _execute_policy_result(self, policy_result: Dict[str, Union[Action, List[Action]]]):
         # If a reset action was given, perform it directly
@@ -556,9 +572,7 @@ class Workflow:
             self._highlight_classifier_result(classifier, cls_name, cls_result, snapshot)
 
         return (
-            [
-                classifier.action(element) for element, _, score in cls_result if score  # type: ignore
-            ]
+            [classifier.action(element) for element, _, score in cls_result if score]  # type: ignore
             if classifier.action
             else []
         )
@@ -568,6 +582,8 @@ class Workflow:
         Resets the workflow.
         Does not clear any history.
         """
+        assert self.config.scraping.history or self.loop_idx == -1, "Cannot reset if config.scraping.history is False!"
+
         self.loop_idx = -1
         self.previous_policy_result = None
 
@@ -591,6 +607,7 @@ class Workflow:
         Because this resets the `loop_idx` variable to match, saved output will override
         previous output.
         """
+        assert self.config.scraping.history, "Cannot reset if config.scraping.history set to False!"
         assert 0 <= view_index < self.loop_idx, "Cannot revert into the future or before the beginning of time!"
         self.reset()
         self.loop_idx = view_index
